# Standard Library
from typing import Any, Dict, List, Optional, Union

# Django
from django.http import HttpRequest

# Local
from .constants import Action
from .renderers import render_turbo_stream
from .response import TurboStreamResponse, TurboStreamTemplateResponse
from .template import render_turbo_stream_template


class TurboStreamTemplate:
    """Wraps template functionality."""

    def __init__(
        self,
        template_name: Union[str, List[str]],
        context: Optional[Dict[str, Any]] = None,
        *,
        action: Action,
        target: str,
        **template_kwargs,
    ):
        self.action = action
        self.target = target
        self.template_name = template_name
        self.context = context
        self.template_kwargs = template_kwargs

    def render(self, **kwargs) -> str:
        return render_turbo_stream_template(
            self.template_name,
            self.context,
            action=self.action,
            target=self.target,
            **{**self.template_kwargs, **kwargs},
        )

    def response(
        self, request: Optional[HttpRequest] = None, **kwargs
    ) -> TurboStreamTemplateResponse:
        request = request or self.template_kwargs.pop("request", None)
        return TurboStreamTemplateResponse(
            request,
            self.template_name,
            self.context,
            action=self.action,
            target=self.target,
            **{**self.template_kwargs, **kwargs},
        )


class TurboStreamAction:
    """Returns strings and responses for a specific Turbo Stream action type."""

    def __init__(self, target: str, action: Action):
        """
        :param target: Turbo Stream target
        :param action: Turbo Stream action
        """
        self.action = action
        self.target = target

    def render(self, content: str = "") -> str:
        """
        :param content: enclosed content
        :return: a *<turbo-stream>* string
        """
        return render_turbo_stream(
            action=self.action, target=self.target, content=content
        )

    def response(self, content: str = "", **response_kwargs) -> TurboStreamResponse:
        """
        :param content: enclosed content
        :return: a *<turbo-stream>* HTTP response wrapper
        """
        return TurboStreamResponse(
            action=self.action, target=self.target, content=content, **response_kwargs
        )

    def template(
        self,
        template_name: Union[str, List[str]],
        context: Optional[Dict[str, Any]] = None,
        **template_kwargs,
    ) -> TurboStreamTemplate:
        """
        :param template_name: Django template name
        :param context: template context

        :return: a *<turbo-stream>* template wrapper
        """
        return TurboStreamTemplate(
            template_name,
            context,
            action=self.action,
            target=self.target,
            **template_kwargs,
        )


class TurboStream:
    """
    Class for creating Turbo Stream strings and responses.
    """

    def __init__(self, target: str):
        """
        :param target: stream target
        """
        self.target = target

    def action(self, action: Action) -> TurboStreamAction:
        return TurboStreamAction(self.target, action)

    @property
    def append(self) -> TurboStreamAction:
        return self.action(Action.APPEND)

    @property
    def prepend(self) -> TurboStreamAction:
        return self.action(Action.PREPEND)

    @property
    def remove(self) -> TurboStreamAction:
        return self.action(Action.REMOVE)

    @property
    def replace(self) -> TurboStreamAction:
        return self.action(Action.REPLACE)

    @property
    def update(self) -> TurboStreamAction:
        return self.action(Action.UPDATE)



# Standard Library
import functools
from typing import Callable, Generator

# Django
from django.http import HttpRequest, HttpResponse

# Local
from .response import TurboStreamResponse, TurboStreamStreamingResponse


def turbo_stream_response(view: Callable) -> Callable:
    @functools.wraps(view)
    def wrapper(request: HttpRequest, *args, **kwargs):
        response = view(request, *args, **kwargs)
        if isinstance(response, HttpResponse):
            return response
        if isinstance(response, Generator):
            return TurboStreamStreamingResponse(response)
        if isinstance(response, str):
            response = list(response)
        return TurboStreamResponse(response)

    return wrapper



# Standard Library
import http
from typing import Any, Dict, Iterable, Mapping, Optional, Union

# Django
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseRedirect,
    StreamingHttpResponse,
)
from django.template.response import TemplateResponse

# Local
from .constants import TURBO_STREAM_MIME_TYPE, Action
from .renderers import render_turbo_frame, render_turbo_stream


class HttpResponseSeeOther(HttpResponseRedirect):
    """Redirect with 303 status"""

    status_code = http.HTTPStatus.SEE_OTHER


class TurboStreamResponseMixin:
    """Automatically sets the correct turbo-stream content type."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            content_type=f"{TURBO_STREAM_MIME_TYPE}; charset=utf-8", *args, **kwargs
        )


class TurboStreamStreamingResponse(TurboStreamResponseMixin, StreamingHttpResponse):
    """Handles turbo-stream streaming responses. Generator should yield individual
    turbo-stream tags.

    For example:

    .. code-block:: python

        def render():

            for i in range(3):

                yield render_turbo_stream(
                    "OK",
                    Action.REPLACE,
                    target=f"item-{i}"
                )

        return TurboStreamStreamingResponse(render())
    """


class TurboStreamResponse(TurboStreamResponseMixin, HttpResponse):
    """Basic turbo-stream response.

    If action and target are provided, will automatically wrap the
    response in <turbo-stream> tags. Otherwise you should provide the
    turbo-stream content string yourself (using e.g. *render_turbo_stream*)
    or an iterable of turbo-streams.
    """

    def __init__(
        self,
        content: Union[Iterable[str], str] = "",
        *,
        action: Optional[Action] = None,
        target: Optional[str] = None,
        **kwargs,
    ):
        if action and target and isinstance(content, str):
            content = render_turbo_stream(action, target, content)
        super().__init__(content, **kwargs)


class TurboStreamTemplateResponse(TurboStreamResponseMixin, TemplateResponse):
    """Handles turbo-stream template response.

    Adds the following variables to the template:

    - **is_turbo_stream**
    - **turbo_stream_action**
    - **turbo_stream_target**

    """

    is_turbo_stream = True

    def __init__(
        self,
        request: HttpRequest,
        template: Union[str, Iterable[str]],
        context: Optional[Dict[str, Any]] = None,
        *,
        action: Action,
        target: str,
        **kwargs,
    ):

        super().__init__(
            request,
            template,
            {
                **(context or {}),
                "turbo_stream_action": action.value,
                "turbo_stream_target": target,
                "is_turbo_stream": True,
            },
            **kwargs,
        )

        self._target = target
        self._action = action

    @property
    def rendered_content(self) -> str:
        return render_turbo_stream(
            action=self._action, target=self._target, content=super().rendered_content
        )


class TurboFrameResponse(HttpResponse):
    """Handles turbo-frame template response."""

    def __init__(self, content: str = "", *, dom_id: str, **kwargs):
        super().__init__(
            render_turbo_frame(dom_id, content),
            **kwargs,
        )


class TurboFrameTemplateResponse(TemplateResponse):
    """Handles turbo-stream template response.

    Adds the following variables to the template:

    - **is_turbo_frame**
    - **turbo_frame_dom_id**

    """

    is_turbo_frame = True

    def __init__(
        self,
        request: HttpRequest,
        template: Union[str, Iterable[str]],
        context: Optional[Mapping[str, Any]] = None,
        *,
        dom_id,
        **kwargs,
    ):

        super().__init__(
            request,
            template,
            {**(context or {}), "turbo_frame_dom_id": dom_id, "is_turbo_frame": True},
            **kwargs,
        )

        self._dom_id = dom_id

    @property
    def rendered_content(self) -> str:
        return render_turbo_frame(self._dom_id, super().rendered_content)



# Standard Library
import http
from typing import Any, Callable, Dict, List, Optional, Type

# Django
from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.http import HttpRequest, HttpResponse
from django.template.engine import Engine
from django.views.generic.edit import FormMixin

# Local
from .constants import Action
from .frame import TurboFrame
from .response import HttpResponseSeeOther
from .stream import TurboStream, TurboStreamAction


class TurboStreamMixin:
    turbo_stream_action: Optional[Action] = None
    turbo_stream_target: Optional[str] = None

    def get_turbo_stream(self) -> TurboStreamAction:

        target = self.get_turbo_stream_target()

        if not target:
            raise ImproperlyConfigured(
                f"turbo stream target not defined in {self.__class__}.get_turbo_stream_target"
            )

        action = self.get_turbo_stream_action()

        if not action:
            raise ImproperlyConfigured(
                f"turbo stream action not defined in {self.__class__}.get_turbo_stream_action"
            )

        return TurboStream(target).action(action)

    def get_turbo_stream_action(self) -> Optional[Action]:
        """Returns the turbo-stream action parameter

        :return: turbo-stream action
        """
        return self.turbo_stream_action

    def get_turbo_stream_target(self) -> Optional[str]:
        """Returns the turbo-stream target parameter

        :return: turbo-stream target
        """
        return self.turbo_stream_target


class TurboFrameMixin:
    turbo_frame_dom_id: Optional[str] = None

    def get_turbo_frame(self) -> TurboFrame:
        dom_id = self.get_turbo_frame_dom_id()
        if not dom_id:
            raise ImproperlyConfigured(
                "turbo frame dom ID not defined in {self.__class__}.get_turbo_frame_dom_id"
            )

        return TurboFrame(dom_id)

    def get_turbo_frame_dom_id(self) -> Optional[str]:
        """Should return a valid DOM ID target for the turbo frame."""
        return self.turbo_frame_dom_id


class TurboStreamResponseMixin(TurboStreamMixin):
    """Handles turbo-stream responses"""

    def get_response_content(self) -> str:
        """Returns turbo-stream content."""

        return ""

    def render_turbo_stream(self, **response_kwargs) -> HttpResponse:
        """Returns a turbo-stream response."""
        return self.get_turbo_stream().response(
            self.get_response_content(), **response_kwargs
        )


class TurboStreamTemplateResponseMixin(TurboStreamMixin):
    """Handles turbo-stream template responses."""

    request: HttpRequest
    template_engine: Engine
    get_template_names: Callable

    def render_turbo_stream(
        self, context: Dict[str, Any], **response_kwargs
    ) -> HttpResponse:
        """Renders a turbo-stream template response.

        :param context: template context
        """

        return (
            self.get_turbo_stream()
            .template(self.get_template_names(), context, using=self.template_engine)
            .response(self.request)
        )


class TurboFormValidationMixin(FormMixin):
    def form_invalid(self, form: forms.Form) -> HttpResponse:
        return self.render_to_response(
            self.get_context_data(form=form),
            status=http.HTTPStatus.UNPROCESSABLE_ENTITY,
        )


class TurboFormMixin(TurboFormValidationMixin, FormMixin):
    """Mixin for handling form validation. Ensures response
    has 422 status on invalid and 303 on success"""

    def form_valid(self, form: forms.Form) -> HttpResponse:
        super().form_valid(form)  # type: ignore
        return HttpResponseSeeOther(self.get_success_url())


class TurboFormAdapterMixin(TurboFormValidationMixin, FormMixin):
    """This is used when you want to adapt an existing view,
    e.g. in 3rd party code."""

    def form_valid(self, form: forms.Form) -> HttpResponse:
        response = super().form_valid(form)  # type: ignore
        # if we already return a response, just change the status code
        if response.status_code in (
            http.HTTPStatus.MOVED_PERMANENTLY,
            http.HTTPStatus.FOUND,
        ):
            response.status_code = http.HTTPStatus.SEE_OTHER
        return response


class TurboFormModelMixin(TurboFormMixin):

    object: Optional[Model]

    def form_valid(self, form: forms.Form) -> HttpResponse:
        """If the form is valid, save the associated model."""
        self.object = form.save()
        return super().form_valid(form)


class TurboStreamFormMixin(TurboStreamMixin, TurboFormMixin):
    """Returns a turbo stream when form is invalid.

    You should define a partial template corresponding to the template
    used to render the initial form, for example if your template
    is `posts/post_form.html` your partial template is `posts/_post_form.html`.

    The template should then be included in the form:

    .. code-block:: html

        {% include "_post_form.html" %}

    The form or container element should have an ID: this is provided as `turbo_stream_target` in
    the template context:


    .. code-block:: html

        <form method="POST" action=".." id="{{ turbo_stream_target }}">

    When the form is posted, if the form is invalid a turbo-stream element wrapping the
    partial content will be returned, including any validation errors.
    """

    turbo_stream_action: Action = Action.REPLACE
    turbo_stream_template_prefix: str = "_"
    turbo_stream_template_name: Optional[str] = None

    template_engine: Engine
    get_template_names: Callable

    def resolve_turbo_stream_template_name(self, template_name):
        """By default template name will have underscore prefix"""
        start, join, name = template_name.rpartition("/")
        return start + join + self.turbo_stream_template_prefix + name

    def get_turbo_stream_template_names(self) -> List[str]:
        if self.turbo_stream_template_name:
            return [self.turbo_stream_template_name]
        return [
            self.resolve_turbo_stream_template_name(template)
            for template in self.get_template_names()
        ]

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        # TurboStream response will automatically add this,
        # but we also want to access it in the initial render
        if "turbo_stream_target" not in kwargs:
            target = self.get_turbo_stream_target()

            if not target:
                raise ImproperlyConfigured(
                    f"turbo stream target not defined in {self.__class__}.get_turbo_stream_target"
                )

            kwargs["turbo_stream_target"] = target

        return super().get_context_data(**kwargs)

    def render_turbo_stream_response(self, **context) -> HttpResponse:

        return (
            self.get_turbo_stream()
            .template(
                self.get_turbo_stream_template_names(),
                context=self.get_context_data(**context),
                using=self.template_engine,
            )
            .response(self.request)
        )

    def form_invalid(self, form: forms.Form) -> HttpResponse:
        return self.render_turbo_stream_response(form=form)


class TurboStreamFormModelMixin(TurboStreamFormMixin):
    object: Optional[Model]
    model: Type[Model]

    def get_turbo_stream_target(self) -> Optional[str]:
        target = super().get_turbo_stream_target()
        if target:
            return target

        if isinstance(self.object, Model):
            return f"form-{self.object._meta.model_name}-{self.object.pk}"

        elif self.model:
            return f"form-{self.model._meta.model_name}"

        return None

    def form_valid(self, form: forms.Form) -> HttpResponse:
        self.object = form.save()
        return super().form_valid(form)


class TurboFrameResponseMixin(TurboFrameMixin):
    """Renders turbo-frame responses"""

    def get_response_content(self) -> str:
        return ""

    def render_turbo_frame(self, **response_kwargs) -> HttpResponse:
        """Renders a turbo frame to response."""

        return self.get_turbo_frame().response(
            self.get_response_content(), **response_kwargs
        )


class TurboFrameTemplateResponseMixin(TurboFrameMixin):
    """Handles turbo-frame template responses."""

    request: HttpRequest
    template_engine: Engine
    get_template_names: Callable

    def render_turbo_frame(
        self, context: Dict[str, Any], **response_kwargs
    ) -> HttpResponse:
        """Returns a turbo-frame response.

        :param context: template context
        """
        return (
            self.get_turbo_frame()
            .template(self.get_template_names(), context, using=self.template_engine)
            .response(self.request)
        )



# Standard Library
import http
from typing import Dict, List, Optional, Union

# Django
from django.db.models import Model
from django.forms import Form
from django.http import HttpRequest
from django.shortcuts import resolve_url
from django.template.response import TemplateResponse

# Local
from .constants import Action
from .response import HttpResponseSeeOther
from .stream import TurboStream


def redirect_303(to: Union[str, Model], *args, **kwargs) -> HttpResponseSeeOther:
    """Sends an HTTP 303 redirect.

    :param to:  URL or view name or model instance. If model then calls `get_absolute_url()`.
    """
    return HttpResponseSeeOther(resolve_url(to, *args, **kwargs))


def render_form_response(
    request: HttpRequest,
    form: Form,
    template: Union[str, List[str]],
    context: Optional[Dict] = None,
    *,
    turbo_stream_template: Optional[Union[str, List[str]]] = None,
    turbo_stream_target: Optional[str] = None,
    turbo_stream_action: Action = Action.REPLACE,
    **response_kwargs,
) -> TemplateResponse:
    """Returns a TemplateResponse with the correct status if the form contains errors.

    If `turbo_stream_template` is provided along with `turbo_stream_target`, a TurboStream
    response will be returned instead if there are form validation errors.
    """

    context = {"form": form, **(context or {})}

    if turbo_stream_template and turbo_stream_target:
        context.update(
            {
                "turbo_stream_target": turbo_stream_target,
                "turbo_stream_template": turbo_stream_template,
            }
        )

        if form.errors:
            return (
                TurboStream(turbo_stream_target)
                .action(turbo_stream_action)
                .template(turbo_stream_template, context)
                .response(request, **response_kwargs)
            )

    return TemplateResponse(
        request,
        template,
        context,
        status=http.HTTPStatus.UNPROCESSABLE_ENTITY
        if form.errors
        else http.HTTPStatus.OK,
        **response_kwargs,
    )



# Standard Library
from typing import Any, Dict, List, Optional, Union

# Django
from django.http import HttpRequest

# Local
from .response import TurboFrameResponse, TurboFrameTemplateResponse, render_turbo_frame
from .template import render_turbo_frame_template


class TurboFrameTemplate:
    """Wraps template functionality."""

    def __init__(
        self,
        template_name: Union[str, List[str]],
        context: Optional[Dict[str, Any]] = None,
        *,
        dom_id: str,
        **template_kwargs,
    ):
        self.template_name = template_name
        self.context = context
        self.template_kwargs = template_kwargs
        self.dom_id = dom_id

    def render(self, **kwargs) -> str:
        """
        :param content: enclosed content
        :return: a *<turbo-frame>* string
        """
        return render_turbo_frame_template(
            self.template_name,
            self.context,
            dom_id=self.dom_id,
            **{**self.template_kwargs, **kwargs},
        )

    def response(
        self, request: Optional[HttpRequest] = None, **kwargs
    ) -> TurboFrameTemplateResponse:
        request = request or self.template_kwargs.pop("request", None)
        return TurboFrameTemplateResponse(
            request,
            self.template_name,
            self.context,
            dom_id=self.dom_id,
            **{**self.template_kwargs, **kwargs},
        )


class TurboFrame:
    """Class for creating Turbo Frame strings and responses."""

    def __init__(self, dom_id: str):
        """
        :param dom_id: DOM ID of turbo frame
        """
        self.dom_id = dom_id

    def render(self, content: str = "") -> str:
        """
        :param content: enclosed content
        :return: a *<turbo-frame>* string
        """
        return render_turbo_frame(dom_id=self.dom_id, content=content)

    def response(self, content: str = "", **response_kwargs) -> TurboFrameResponse:
        """
        :param content: enclosed content
        :return: a *<turbo-frame>* HTTP response
        """
        return TurboFrameResponse(
            dom_id=self.dom_id, content=content, **response_kwargs
        )

    def template(
        self,
        template_name: Union[str, List[str]],
        context=Optional[Dict[str, Any]],
        **template_kwargs,
    ) -> TurboFrameTemplate:
        """
        :param template_name: Django template name(s)
        :param context: template context
        :return: a *<turbo-frame>* HTTP response
        """
        return TurboFrameTemplate(
            template_name, context, dom_id=self.dom_id, **template_kwargs
        )



# Standard Library
from typing import Callable

# Django
from django.http import HttpRequest, HttpResponse
from django.utils.functional import SimpleLazyObject

# Local
from .constants import TURBO_STREAM_MIME_TYPE


class TurboData:
    def __init__(self, request: HttpRequest):
        self.has_turbo_header = TURBO_STREAM_MIME_TYPE in request.headers.get(
            "Accept", ""
        )
        self.frame = request.headers.get("Turbo-Frame", None)

    def __bool__(self):
        return self.has_turbo_header


class TurboMiddleware:
    """Adds `turbo` attribute to request:
    1. `request.turbo` : True if request contains turbo header
    2. `request.turbo.frame`: DOM ID of requested Turbo-Frame (or None)
    """

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]):
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:

        request.turbo = SimpleLazyObject(lambda: TurboData(request))
        return self.get_response(request)
